#include "ipv4.h"

#ifndef WINDOWS
#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"
#else
#define ANSI_COLOR_RED     ""
#define ANSI_COLOR_GREEN   ""
#define ANSI_COLOR_YELLOW  ""
#define ANSI_COLOR_BLUE    ""
#define ANSI_COLOR_MAGENTA ""
#define ANSI_COLOR_CYAN    ""
#define ANSI_COLOR_RESET   ""
#endif

char * ipntoa
(const uint32_t ip, char * dst, const int length)
{
  unsigned char oct[4];

  if (length < 16)
    {
      fprintf(stderr, "ipntoa: buffer must be at least 16 bits wide.\n");
      exit(-1);
    }

  if (dst == NULL)
    {
      dst = malloc(sizeof(char) * 16);
    }

  memset(dst, 0, length);
  oct[0] = (unsigned char)(ip >> 24);
  oct[1] = (unsigned char)(ip >> 16);
  oct[2] = (unsigned char)(ip >> 8);
  oct[3] = (unsigned char)(ip & 0x000000ff);

  if (snprintf(dst, length, "%d.%d.%d.%d", oct[0], oct[1], oct[2], oct[3]) == length)
    {
      fprintf(stderr, "Warning: ipntoa exceeded buffer size.\n");
    }
  return(dst);
}

uint32_t networkNumber
(uint32_t ip, uint32_t sm)
{
    return (ip & sm);
}

uint32_t subnetDirectedBroadcast
(uint32_t ip, uint32_t sm)
{
    return (ip | ~sm);
}

uint32_t wildcardMask
(uint32_t sm)
{
    return ~sm;
}

char classful
(const uint32_t ip)
{
    char class = 0;
    if (ip >= 0xf0000000)      /* High bits = 1111 */
        class = 'E';
    else if (ip >= 0xe0000000) /* High bits = 1110 */
        class = 'D';
    else if (ip >= 0xc0000000) /* High bits = 110  */
        class = 'C';
    else if (ip >= 0x80000000) /* High bits = 10   */
        class = 'B';
    else                       /* High bit  = 0    */
        class = 'A';
    return(class);
}

uint32_t classfulMask
(const uint32_t ip)
{
    uint32_t mask = 0;
    char s[16];

    switch (classful(ip))
      {
      case 'A': mask = 0xff000000; break;
      case 'B': mask = 0xffff0000; break;
      case 'C': mask = 0xffffff00; break;
      case 'D':
      case 'E':
	ipntoa(ip, s, 16);
	fprintf(stderr,
		ANSI_COLOR_RED
		"%s is not a routable unicast IP and therefore has no "
		"classful subnet mask.\n"
		ANSI_COLOR_RESET, s);	
      }
    return(mask);
}

uint32_t summaryAddress
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
  return (networkNumber(ip1, sm1) & networkNumber(ip2, sm2) 
	  & summaryPrefix(ip1, sm1, ip2, sm2));
}

uint32_t summaryPrefix
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
  /* The Common Network Mask is the bits shared by each IP network. */
  uint32_t cnm;

  /* This is the Prefix that summarizes the two networks.
     It can be equal to or shorter than either input subnet mask. */
  uint32_t pfx;

  /* These are the bits different in each network number. */
  uint32_t diff;

  uint32_t network1, network2;

#ifdef DEBUG_SUMMARY_PREFIX
  char diff_s [16];
  char cnm_s [16];
  char pfx_s [16];
  char sm_s [16];
#endif

  network1 = networkNumber(ip1, sm1);
  network2 = networkNumber(ip2, sm2);

  /* Extra & against opposite subnet mask discovers
      if one network is a subnet of the other,
      or the same network. */
  if ((network1 & sm2) == (network2 & sm1))
    {
      pfx = sm1 & sm2;
    }
  else
    {
      diff = network1 ^ network2;
      /* Possible bugs here - clz can produce unexpected results. */
      cnm = 0xffffffff << (32 - __builtin_clz(diff));
      if (cnm > (sm1 & sm2))
	{
	  fprintf(stderr, ANSI_COLOR_RED);
	  fprintf(stderr,
		  "QA Warning: this line should be unreachable. "
		  "The LZCNT instruction may have been used but not "
		  "supported.\n");
	  fprintf(stderr, ANSI_COLOR_RESET);
	  pfx = (sm1 & sm2);
	}
      else
	{
	  pfx = cnm;
	}
    }

#ifdef DEBUG_SUMMARY_PREFIX
  fprintf(stderr, ANSI_COLOR_BLUE);
  fprintf(stderr, "                 SUMMARY PREFIX\n");
  fprintf(stderr, "diff = %-16s\tclz(diff) = %d\n",
	  ipntoa(diff, diff_s, 16), __builtin_clz(diff));
  fprintf(stderr, "cnm  = %-16s\tsm1 & sm2 = %s\n",
	  ipntoa(cnm, cnm_s, 16), ipntoa(sm1 & sm2, sm_s, 16));
  fprintf(stderr, "pfx  = %-16s\n", ipntoa(pfx, pfx_s, 16));
  fprintf(stderr, ANSI_COLOR_RESET "\n");
#endif

  return pfx;
}

uint32_t prefixLength
(const uint32_t sm)
{
  return ((33 - __builtin_ffs(sm)) % 33); /* mod 33 to handle special 0 case */
}

uint32_t countHosts
(const uint32_t sm)
{
  uint32_t prefix, hosts;
  prefix = prefixLength(sm);
  if (sm > 0x00000000 && sm < 0xfffffffe) /* 1.0.0.0 - 255.255.255.252 */
    {
      hosts = (1 << (32 - prefix)) - 2;
    }
  else if (sm == 0xfffffffe)              /* 255.255.255.254 */
    {

    }
  else                                    /* 0.0.0.0 and 255.255.255.255 */
    {
      hosts = 0;
    }
  return(hosts);
}

uint32_t countSubnets
(const uint32_t ip, const uint32_t sm)
{
    uint32_t subnets = 0;
    uint32_t classful = classfulMask(ip);

    if (sm > classful)
    {
        subnets = (1 << (prefixLength(sm) - prefixLength(classful)));
    }

    return(subnets);
}

int isValidNetmask
(const uint32_t sm)
{
  uint32_t valid = 0;
  if (sm == 0)
    {
      fprintf(stderr,
	      "0.0.0.0 may not technically be a valid subnet mask.\n");
      valid = 1;
    }
  else if (__builtin_ffs(sm) + __builtin_popcount(sm) - 1 == 32)
    {
      valid = 1;
    }
  else
    {
      char * str = calloc(sizeof(char), 16);
      fprintf(stderr,
	      ANSI_COLOR_RED
	      "%s is not a valid subnet mask. FFS=%-2d POP=%-2d\n"
	      ANSI_COLOR_RESET,
	      ipntoa(sm, str, 16),
	      __builtin_ffs(sm) - 1,
	      __builtin_popcount(sm));
      free(str);
      exit(-1);
    }
  return (valid);
}

uint32_t parse
(const char * src)
{
  uint32_t ip, i, x;
  unsigned int a, dotCount;
  char * token;
  char s[16];
  const char delim[2] = ".";

  strncpy(s, src, 16);

  /* Let's do several sanity checks before doing anything. */
  if (strlen(s) > 16)
    {
      fprintf(stderr, ANSI_COLOR_RED "%s is not a valid IP.\n"
	      ANSI_COLOR_RESET, s);
      exit(-1);
    }

  for (a = 0 ; a < strlen(src) ; a++)
    {
      if (s[a] != '.' && (s[a] < '0' || s[a] > '9'))
	{
	  fprintf(stderr, ANSI_COLOR_RED
		  "%s is not a valid IP (invalid character: %c).\n"
		  ANSI_COLOR_RESET, src, s[a]);
	  exit(-1);
	}
    }

  if (src[0] == '.' || src[strlen(src) - 1] == '.')
    {
      fprintf(stderr, ANSI_COLOR_RED
	      "%s begins or ends with a '.'.\n"
	      ANSI_COLOR_RESET, src);
      exit(-1);
    }

  dotCount = 0;
  for (a = 0 ; a < strlen(src) ; a++)
    {
      if (src[a] == '.')
	{
	  dotCount++;
	}
    }
  if (dotCount != 3)
    {
      fprintf(stderr, ANSI_COLOR_RED
	      "%s does not contain three periods and is not a valid "
	      "IPv4 address.\n"
	      ANSI_COLOR_RESET, src);
      exit(-1);
    }

  ip = 0;
  i = 32;

  token = strtok(s, delim);
  while (i > 0 && token != NULL)
    {
      i -= 8;
      x = (uint32_t) atoi(token);
      
      if (x > 255)
	{
	  fprintf(stderr, ANSI_COLOR_RED
		  "%s is not a valid IP (invalid value: %s)\n"
		  ANSI_COLOR_RESET, src, token);
	  exit(-1);
	} 

      ip += (x << i);
      token = strtok(NULL, delim);
    }
  return(ip);
}

void generateSubnets
()
{
  uint32_t mask, i;
  char ip[16];
  mask = 0x00000000;

  /* though ugly, 33 iterations ensures we catch case 0xffffffff */
  for (i=0; i<33; i++)
    {
      ipntoa(mask, ip, 16);
      printf("%-2d\t%s\n", prefixLength(mask), ip);
      mask = (mask >> 1) + 0x80000000;
    }
}
