#include "ipv4.h"

#ifndef WINDOWS
#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"
#else
#define ANSI_COLOR_RED     ""
#define ANSI_COLOR_GREEN   ""
#define ANSI_COLOR_YELLOW  ""
#define ANSI_COLOR_BLUE    ""
#define ANSI_COLOR_MAGENTA ""
#define ANSI_COLOR_CYAN    ""
#define ANSI_COLOR_RESET   ""
#endif

const char * usage = "Usage: ./ipv4 [ip1 sm1] {ip2 sm2} | [generate subnets] | [classful]\n";

void testNetworkNumber
(const char * ip1, const char * sm1,
 const char * ip2, const char * sm2)
{
    uint32_t ip1_i, ip2_i, sm1_i, sm2_i;
    char network1[16], network2[16];

    ip1_i = parse(ip1);
    sm1_i = parse(sm1);
    isValidNetmask(sm1_i);
    ipntoa(networkNumber(ip1_i, sm1_i), network1, 16);

    if (ip2 != NULL && sm2 != NULL)
      {
	ip2_i = parse(ip2);
	sm2_i = parse(sm2);
	isValidNetmask(sm2_i);
	ipntoa(networkNumber(ip2_i, sm2_i), network2, 16);
      }

    printf("                NETWORK NUMBERS\n");
    printf("IP                Subnet Mask       Prefix  Network Number   Class\n");
    printf("%-15s   %-15s   %-2d      %-15s  %c\n", ip1, sm1, prefixLength(sm1_i), network1, classful(ip1_i));
    if (ip2 != NULL && sm2 != NULL)
      printf("%-15s   %-15s   %-2d      %-15s  %c\n", ip2, sm2, prefixLength(sm2_i), network2, classful(ip2_i));
}

void testSubnet
(const char * ip1, const char * sm1,
 const char * ip2, const char * sm2)
{
  uint32_t ip1_i, ip2_i, sm1_i, sm2_i, bc1, bc2;
  char * broadcast1, * broadcast2;

    broadcast1 = calloc(sizeof(char), 16);
    broadcast2 = calloc(sizeof(char), 16);

    ip1_i = parse(ip1);
    sm1_i = parse(sm1);
    bc1 = subnetDirectedBroadcast(ip1_i, sm1_i);
    ipntoa(bc1, broadcast1, 16);

    if (ip2 != NULL && sm2 != NULL)
      {
	ip2_i = parse(ip2);
	sm2_i = parse(sm2);
	bc2 = subnetDirectedBroadcast(ip2_i, sm2_i);
	ipntoa(bc2, broadcast2, 16);
      }

    printf("             SUBNET DIRECTED BROADCAST\n");
    printf("IP                Subnet Mask       Prefix  Broadcast\n");
    printf("%-15s   %-15s   %-2d      %-15s\n", ip1, sm1, prefixLength(sm1_i), broadcast1);
    if (ip2 != NULL && sm2 != NULL)
      printf("%-15s   %-15s   %-2d      %-15s\n", ip2, sm2, prefixLength(sm2_i), broadcast2);

    free(broadcast1);
    free(broadcast2);
}

void testNetworkSize
(const char * ip1, const char * sm1,
 const char * ip2, const char * sm2)
{
  uint32_t ip1_i, ip2_i, sm1_i, sm2_i, net1_i, net2_i;
  char cidr1[20], cidr2[20], net1_s[16], net2_s[16];

  ip1_i = parse(ip1);
  sm1_i = parse(sm1);
  net1_i = networkNumber(ip1_i, sm1_i);
  ipntoa(net1_i, net1_s, 16);
  snprintf(cidr1, 20, "%s/%d", net1_s, prefixLength(sm1_i));

  if (ip2 != NULL && sm2 != NULL)
    {
      ip2_i = parse(ip2);
      sm2_i = parse(sm2);
      net2_i = networkNumber(ip2_i, sm2_i);
      ipntoa(net2_i, net2_s, 16);
      snprintf(cidr2, 20, "%s/%d", net2_s, prefixLength(sm2_i));
    }

  printf("                  NETWORK SIZE\n");
  printf("Network             Subnets\tHosts\n");
  printf("%-18s  %-7d\t%d\n", cidr1, countSubnets(ip1_i, sm1_i), countHosts(sm1_i));
  if (ip2 != NULL && sm2 != NULL)
    printf("%-18s  %-7d\t%d\n", cidr2, countSubnets(ip2_i, sm2_i), countHosts(sm2_i));
}

void testSummary
(const char * ip1, const char * sm1,
 const char * ip2, const char * sm2)
{
  uint32_t ip1_i, ip2_i, sm1_i, sm2_i, sumpfx;
    char * sip, * ssm, * wcm;

    ip1_i = parse(ip1);
    ip2_i = parse(ip2);
    sm1_i = parse(sm1);
    sm2_i = parse(sm2);

    sip = calloc(sizeof(char), 16);
    ssm = calloc(sizeof(char), 16);
    wcm = calloc(sizeof(char), 16);

    sumpfx = summaryPrefix(ip1_i, sm1_i, ip2_i, sm2_i);

    ipntoa(summaryAddress(ip1_i, sm1_i, ip2_i, sm2_i), sip, 16);
    ipntoa(sumpfx, ssm, 16);
    ipntoa(wildcardMask(sumpfx), wcm, 16);

    printf("                 SUMMARY ROUTE\n");
    printf("Summary IP        Summary Mask      Prefix   Wildcard Mask\n");
    printf("%-15s   %-15s   %-2d       %-15s\n", sip, ssm, prefixLength(sumpfx), wcm);
    free(sip);
    free(ssm);
    free(wcm);
}

void testClassfulAddresses()
{
#define CLASSFUL_TESTS 5
  unsigned short i = 0;
  uint32_t ip[CLASSFUL_TESTS];
  char ip_s[16], sm_s[16];
  uint32_t sm = 0;

  ip[0] = parse("10.0.0.1");
  ip[1] = parse("172.16.0.1");
  ip[2] = parse("192.168.0.1");
  ip[3] = parse("224.0.0.1");
  ip[4] = parse("255.255.255.255");

  printf("C  IP               Mask\n");
  while (i < CLASSFUL_TESTS)
    {
      char class;
      ipntoa(ip[i], ip_s, 16);
      class = classful(ip[i]);
      if (class <= 'C')
	{
	  sm = classfulMask(ip[i]);
	  ipntoa(sm, sm_s, 16);
	  printf("%c  %-15s  %-15s\n", class, ip_s, sm_s);
	}
      else
	{
	  printf("%c  %-15s\n", class, ip_s);
	}
      i++;
    }
}

void testAll
(const char * ip1, const char * sm1,
 const char * ip2, const char * sm2)
{
    testNetworkNumber(ip1, sm1, ip2, sm2);
    printf("\n");
    testSubnet(ip1, sm1, ip2, sm2);
    printf("\n");
    testNetworkSize(ip1, sm1, ip2, sm2);
    printf("\n");
    testSummary(ip1, sm1, ip2, sm2);
}

int main
(int argc, char ** argv)
{
  int i;
  for (i = 1; i < argc; i++)
    {
      if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
	{
	  fprintf(stderr, ANSI_COLOR_YELLOW "%s\n" ANSI_COLOR_RESET, usage);
	  return(EXIT_SUCCESS);
	}
    }

    if (argc == 5)
    {
      testAll(argv[1], argv[2], argv[3], argv[4]);
    }
    else if (argc== 3 && (strncmp(argv[1], "generate", 10) == 0) && 
	     (strncmp(argv[2], "subnets", 8) == 0))
      {
	generateSubnets();
      }
    else if (argc == 2 && (strncmp(argv[1], "classful", 8) == 0))
      {
	testClassfulAddresses();
      }
    else if (argc == 3)
      {
	testNetworkNumber(argv[1], argv[2], NULL, NULL);
	printf("\n");
	testSubnet(argv[1], argv[2], NULL, NULL);
	printf("\n");
	testNetworkSize(argv[1], argv[2], NULL, NULL);
      }
    else
    {
      fprintf(stderr, ANSI_COLOR_YELLOW "%s\n" ANSI_COLOR_RESET, usage);
    }

    return(EXIT_SUCCESS);
}
